<?php

/**
 * @file build.instance.inc
 * Build an instance using a site alias
 */

/**
 * Implements drush_COMMAND_validate().
 */
function drush_instance_build_validate($alias) {
  // Check alias exists
  if(!($alias_info = drush_sitealias_get_record($alias))) {
    drush_set_error("DRUSH_INSTANCE_SITEALIAS_NONE", "Alias $alias does not exist");
  }
  // Check it's a local alias
  elseif (array_key_exists('remote-host', $alias_info)) {
    drush_set_error("DRUSH_INSTANCE_SITEALIAS_REMOTE", "Sorry, drush instance does not currently support remote hosts like $alias.");
  }
  // Check it has a source for the makefile
  elseif (!(array_key_exists('instance', $alias_info)
            && array_key_exists('sources', $alias_info['instance'])
            && array_key_exists('makefile', $alias_info['instance']['sources'])
         )) {
    drush_set_error("DRUSH_INSTANCE_GIT_REPO_MAKEFILE", "No makefile source specified for $alias: you should specify \$alias['instance']['sources']['makefile'] as a makefile-like definition");
  }

  // Add the build/rebuild status
  $alias_info['_is_rebuild'] = drush_get_option('rebuild');
  // Add the option for which useful templates to build
  $alias_info['_has_templates'] = drush_get_option('templates');
  // Unset these and any other options, so drush make doesn't pick up on them
  $alias_info['command_line'] = array();
  $commands = instance_drush_command();
  foreach(array_keys($commands['instance-build']['options']) as $option) {
    $alias_info['command_line'][$option] = drush_get_option($option);
    drush_unset_option($option);
  }

  // Check whether or not the site exists
  if (file_exists($alias_info['root'])) {
    // This is OK, if rebuild is specified
    if (!$alias_info['_is_rebuild']) {
      drush_set_error("DRUSH_INSTANCE_ALREADY_EXISTS", "Instance $alias already exists at " . $alias_info['root']);
    }
  }
  // If rebuild IS specified, but there's no existing build, raise error
  elseif ($alias_info['_is_rebuild']) {
    drush_set_error("DRUSH_INSTANCE_ALREADY_EXISTS", "Instance $alias does not exist at " . $alias_info['root'] . "; cannot rebuild");
  }

  // If there's an error by this point, we shouldn't continue
  if (drush_get_error() == DRUSH_FRAMEWORK_ERROR) {
    return;
  }

  // Makefile-as-version-control branch/tag/revision override
  // Refspec types copied in priority order from make_download_git()
  $all_refspec_types = array('tag', 'revision', 'branch');
  foreach($all_refspec_types as $refspec_type) {
    if ($refspec = $alias_info['command_line'][$refspec_type]) {
      $alias_info['instance']['sources']['makefile'][$refspec_type]
        = $refspec;
      // Unset every other type, so the command-line override prevails
      foreach($all_refspec_types as $unset_type) {
        if ($unset_type != $refspec_type) {
          unset($alias_info['instance']['sources']['makefile'][$unset_type]);
        }
      }
      // Once we've found one, we're done
      break;
    }
  }

  // Defaults
  // Makefile name
  if (!array_key_exists('makefile_path', $alias_info['instance'])) {
    $alias_info['instance']['makefile_path'] = 'instance.make';
  }
  // If there's no extra source for sites/default, assume it's the same as the makefile
  if (!array_key_exists('sites_default', $alias_info['instance']['sources'])) {
    $alias_info['instance']['sources']['sites_default'] =
      $alias_info['instance']['sources']['makefile'];
  }

  // Validation of paths array
  // used to be only run during rebuild
  _drush_instance_build_validate_paths($alias_info);

  // Save the site alias in context for use in the actual command
  drush_set_context("DRUSH_INSTANCE", $alias_info);
}

/**
 * Implements drush_hook().
 */
function drush_instance_build($alias) {
  // Get all the config we'll need first
  $alias_info = drush_get_context("DRUSH_INSTANCE");
  $instance = $alias_info['instance'];
  $rebuild = (bool)$alias_info['_is_rebuild'];

  // Try to find makefile
  drush_log("Looking for makefile", "ok");
  $master_tmp = make_tmp();
  $makefile_tmp = "$master_tmp/makefile";
  _drush_instance_make_download_factory(
    'Instance-makefile',
    $instance['sources']['makefile'],
    $makefile_tmp
  );
  $makefile_location = "$makefile_tmp/" . $instance['makefile_path'];
  if(!file_exists($makefile_location)) {
    return drush_set_error("DRUSH_INSTANCE_MAKEFILE_NONE", "Makefile not present in source.");
  }

  // Make site
  drush_log("Making site", "ok");
  // Drush make uses make_tmp() and doesn't play nicely (directory name
  // collisions, and it seems it deletes everything when it's done)
  // So generate our own temporary location
  $site_tmp = _drush_instance_tmp_unique();

  // Run drush make
  drush_invoke('make', array($makefile_location, $site_tmp));

  // Replace sites/default with source specified, defaulting to
  // the makefile source (see validation hook)
  $sites_default = "$site_tmp/sites/default";
  drush_delete_dir($sites_default);
  // Assume deployed codebase should always be a working copy, so hot-swap a
  // TRUE value into that drush_get_option before downloading from source
  $old_wc = drush_get_option('working-copy');
  drush_set_option('working-copy', TRUE);
  _drush_instance_make_download_factory(
    'Sites-default',
    $instance['sources']['sites_default'],
    $sites_default
  );
  // Swap the old option for working-copy back in
  drush_set_option('working-copy', $old_wc);
  drush_log("Replaced sites/default with relevant source", "ok");

  // Finally, move it to the site alias root
  // If this is a rebuild, then there will be an existing file root
  if($rebuild) {
    // Move it to one side and report success or failure
    $old_dir = $alias_info['root'] . '-old-' . rand(1, 99999);
    $alias_info['_old_dir'] = $old_dir;
    $ret = drush_move_dir($alias_info['root'], $old_dir);
    if (!$ret) {
      return drush_set_error("Cannot move existing instance aside in order to rebuild");
    }
    drush_log("Moved existing instance aside to $old_dir", "ok");
  }

  // Move new build into place
  $ret = drush_move_dir($site_tmp, $alias_info['root'], FALSE);
  if (!$ret) {
    return drush_set_error("Cannot move instance into site alias location");
  }
  drush_log("Moved new instance in place at " . $alias_info['root'], "ok");

  // Now move all preserved files over: we do this as late
  // as possible to avoid e.g. moving large filetrees across partitions,
  // as we can assume that e.g. /var/www/* is all on one partition (?)
  // but the same can't be said for e.g. /tmp
  // Also change permissions if required at this point
  $path_success = TRUE;
  foreach($instance['paths'] as $path => $config) {
    // Move file
    $path_success = $path_success && _drush_instance_build_path($alias_info, $path, $config, $rebuild);
    if (!$path_success) {
      drush_log("Could not move into new build: $path", "warning");
      break;
    }
    // Change/set permissions?
    if (array_key_exists("permissions", $config)) {
      $path_success = $path_success && _drush_instance_perm_path($alias_info, $path, $config);
    }
    if (!$path_success) {
      drush_log("Could not set permissions: $path", "warning");
      break;
    }
  }
  if (!$path_success) {
    return drush_set_error("Your codebase has been rebuilt, but some paths have had problems. It could therefore be in a partially built state, so please check the error log above.");
  }

  // Build some template files
  if ($alias_info['_has_templates']) {
    $templates = split(",",
      $alias_info['_has_templates'] == "all" ? "vhost,crontab" : $alias_info['_has_templates']);
    foreach($templates as $template_key) {
      _drush_instance_render_to_file($template_key, substr($alias, 1), $alias_info);
    }
  }

  // Finally, we're done!
  drush_log(dt("Finished !alias: deployed to !root",
            array("!alias" => $alias, "!root" => $alias_info['root'])),
            "ok");
}

/**
 * Generic handler for all paths
 */
function _drush_instance_build_path($alias_info, $path, $config, $rebuild) {
  // Specific handler based on type, or 'rebuild' if that's appropriate
  $type = $rebuild ? 'rebuild' : $config['source']['type'];
  $handler = "_drush_instance_build_path_$type";
  if (!function_exists($handler)) {
    drush_set_error("No handler method exists for preserving paths of type=$type");
    return FALSE;
  }
  return $handler($alias_info, $path, $config);
}

/**
 * Specific handler for rebuild paths
 */
function _drush_instance_build_path_rebuild($alias_info, $path, $config) {
  // Only preserve paths marked as such
  if (!$config['preserve_on_rebuild']) {
    return TRUE;
  }

  // Shortcut variables for old and new paths
  $full_old_path = $alias_info['_old_dir'] . "/$path";
  $full_new_path = $alias_info['root'] . "/$path";

  // Ensure all directories above this one exist
  if(!_drush_instance_build_path_containing_dir($full_new_path)) {
    drush_log("Could not create all subdirectories required to preserve $path", "warning");
    return FALSE;
  }

  // Again, to avoid replicating large files, we should use rename() -
  // drush_move_dir() actually copies the contents
  $ret = @drush_op('rename', $full_old_path, $full_new_path);
  if (!$ret) {
    drush_log("Could not move $path to new build", "warning");
  }
  else {
    drush_log("Moved $path to new build", "ok");
  }
  return $ret;
}

/**
 * Specific handler for touch paths
 */
function _drush_instance_build_path_touch($alias_info, $path, $config) {
  // Shortcut variable for new path
  $full_new_path = $alias_info['root'] . "/$path";

  // Ensure all directories above this one exist
  if(!_drush_instance_build_path_containing_dir($full_new_path)) {
    drush_log("Could not create all subdirectories required to set up empty file $path", "warning");
    return FALSE;
  }

  return (@file_put_contents($full_new_path, '') !== FALSE);
}

/**
 * Specific handler for touch paths
 */
function _drush_instance_build_path_mkdir($alias_info, $path, $config) {
  // Shortcut variable for new path
  $full_new_path = $alias_info['root'] . "/$path";

  // Ensure all directories above this one exist
  if(!_drush_instance_build_path_containing_dir($full_new_path)) {
    drush_log("Could not create all subdirectories required to set up directory $path", "warning");
    return FALSE;
  }

  return @drush_mkdir($full_new_path);
}

/**
 * Specific handler for copy paths
 */
function _drush_instance_build_path_copy($alias_info, $path, $config) {
  // Shortcut variable for new path
  $full_new_path = $alias_info['root'] . "/$path";

  // Ensure all directories above this one exist
  if(!_drush_instance_build_path_containing_dir($full_new_path)) {
    drush_log("Could not create all subdirectories required to copy $path", "warning");
    return FALSE;
  }

  // Copy using copy() or drush's custom command, depending on whether or not
  // the source is a file or directory
  $full_from_path = $alias_info['root'] . "/" . $config['source']['from'];
  if (is_dir($full_from_path)) {
    return drush_copy_dir($full_from_path, $full_new_path);
  }
  return copy($full_from_path, $full_new_path);
}

/**
 * Specific handler for rsync paths
 */
function _drush_instance_build_path_rsync($alias_info, $path, $config) {
  // Shortcut variable for new path
  $full_new_path = $alias_info['root'] . "/$path";

  // Assemble any options for drush rsync, including the rsync source
  if (!array_key_exists('from', $config['source'])) {
    return drush_set_error("Path source of type=rsync needs a 'from' specification.");
  }
  $from = $config['source']['from'];
  $additional_options = $config['source'];
  unset($additional_options['type']);
  unset($additional_options['from']);

  // Ensure all directories above this one exist
  if(!_drush_instance_build_path_containing_dir($full_new_path)) {
    drush_log("Could not create all subdirectories required to set up directory $path", "warning");
    return FALSE;
  }

  // Run rsync
  drush_command_include('core-rsync');
  include_once(dirname(__FILE__) . "/forked.rsync.inc");
  // Swap out DRUSH_COMMAND_ARGS context
  // Note that it might also need to swap out other contexts in future for
  // e.g. default arguments
  $context =& drush_get_context();
  $cached_args = $context['DRUSH_COMMAND_ARGS'];
  $context['DRUSH_COMMAND_ARGS'] = array();
  // Ensure the target directory exists
  @drush_mkdir($full_new_path);
  // Appending "/" to both paths ensures directories get rsynced, not
  // a new subdirectory created locally
  $return_value = _instance_drush_core_rsync($from . "/", $full_new_path . "/", $additional_options);

  // Swap context back in and return
  $context['DRUSH_COMMAND_ARGS'] = $cached_args;
  return $return_value;
}

/**
 * Specific handler for symlink paths
 */
function _drush_instance_build_path_symlink($alias_info, $path, $config) {
  // Shortcut variable for new path
  $to_path = $alias_info['root'] . "/$path";
  // Path we're linking FROM
  $from_path = $config['source']['from'];
  // Make absolute
  if (strpos($from_path, "/") !== 0) {
    $from_path = $alias_info['root'] . "/$from_path";
  }

  // Ensure all directories above this one exist
  if(!_drush_instance_build_path_containing_dir($to_path)) {
    drush_log("Could not create all subdirectories required to set up symlink $path", "warning");
    return FALSE;
  }

  return symlink($from_path, $to_path);
}

/**
 * Helper function to create all the directories down to a new path, but
 * not including it
 */
function _drush_instance_build_path_containing_dir($full_new_path) {
  // Check all directories on new path exist, not including the last bit
  $check_path = "/";
  foreach(split("/", $full_new_path) as $new_path_part) {
    if(!file_exists($check_path) && !drush_mkdir($check_path)) {
      return FALSE;
    }
    $check_path .= "/$new_path_part";
  }
  return TRUE;
}

/**
 * Generic handler for all permissions
 */
function _drush_instance_perm_path($alias_info, $path, $config) {
  // Specific handler based on type
  $type = $config['permissions']['type'];
  $handler = "_drush_instance_perm_path_$type";
  if (!function_exists($handler)) {
    drush_set_error("No handler method exists for setting perms on paths of type=$type");
    return FALSE;
  }
  return $handler($alias_info, $path, $config['permissions']);

}

/**
 * Helper function to make a file writeable by all
 */
function _drush_instance_perm_path_chmod_all($alias_info, $path, $config) {
  return chmod($alias_info['root'] . "/$path", 0757);
}

/**
 * Helper function to make a file writeable by group
 */
function _drush_instance_perm_path_chmod_group($alias_info, $path, $config) {
  return chmod($alias_info['root'] . "/$path", 0775);
}

/**
 * Stub-out of theme_render_template().
 *
 * Render templates into usable configuration files
 */
function _drush_instance_theme_render_template($template_file, $variables) {
  extract($variables, EXTR_SKIP); // Extract the variables to a local namespace
  ob_start(); // Start output buffering
  include dirname(__FILE__) . "/$template_file"; // Include the template file
  $contents = ob_get_contents(); // Get the contents of the buffer
  ob_end_clean(); // End buffering and discard
  return $contents; // Return the contents
}

/**
 * Render one of our template files using the above
 */
function _drush_instance_render_to_file($key, $alias_name, $alias_info) {
  // Does template file exist
  $file = "templates/$key.tpl.php";
  if (!file_exists(dirname(__FILE__) . "/$file")) {
    return drush_log("Template $key cannot be rendered as it doesn't exist", "warning");
  }
  // Render template to string
  $content = _drush_instance_theme_render_template(
    $file,
    array("alias_name" => substr($alias, 1), "alias_info" => $alias_info)
  );

  file_put_contents($alias_info['root'] . "/sites/$key", $content);
}
